
* Bend.gms:  Benders decomposition algorithm.
* Nielsen/Zenios:  A Library of Financial Optimization Models, Section 5.4.
* Last modified: Oct 2002.
*
*
* This version: x is first-stage (complicating), y is second-stage.
*
* The code solves the following generic problem:
*
*          Min z = c^t x + d^t y
*          S.t Ax + By >= rhs;
*          0 <= x <= u_x, 0 <= y <= u_y
*
* Dimensions: A is m by n1, B is m by n2.
* Indeyed by:      i,   j1,      i,   j2.
*
* The subproblems use a formulation with a penalty variable,
* p, with a large objective coefficient, BigM, to ensure sub-
* problem feasibility.
*   Whenever the subproblem is "really" infeasible, divide the 
* cut by BigM to see what the feasibility cut would have looked like!
*
$eolcom #
*
* First formulate and solve the problem as a one-shot.
*
  SET m, n1, n2;
  ALIAS (m, i), (n1, j1), (n2, j2);

  PARAMETER A(i,j1), B(i,j2), c(j1), d(j2), rhs(i), ux(j1), uy(j2);

* Algorithm parameters (max be changed in include file if needed):
  SCALAR BigM /   10000 /;   # Obj. Coeffient for penaltx variable p
  SCALAR Epsilon / 0.01 /;   # Termination tolerance.

  OPTION LP = CPLEX;

$include "GarNem.inc"

* Set up the main problem:
  EQUATIONS MainObj, MainCon(i);
  VARIABLE z, x(j1), y(j2);

  MainObj      .. z =E= sum( j1, c(j1) * x(j1) ) + 
                        sum( j2, d(j2) * y(j2) );

  MainCon(i)   ..       sum( j1, A(i,j1) * x(j1) ) + 
                        sum( j2, B(i,j2) * y(j2) ) =G= rhs(i);

  x.LO(j1) = 0;      y.LO(j2) = 0;
  x.UP(j1) = ux(j1); y.UP(j2) = uy(j2);

  MODEL MainModel /MainObj, MainCon/;
  SOLVE MainModel MINIMIZING z USING LP;

************** END SOLVING MAIN PROBLEM. ************************************



************** SET UP BENDERS DECOMPOSITION *********************************

* 
* Iteration control:
*
  SET Iter / BendIt_1 * BendIt_1000 /; ALIAS(Iter, It); 
  SET Cut(Iter);      # Benders iterations alreadx performed
  SET this(Iter);     # Current Benders iteration

  PARAMETER y_sol(iter, j2), pi(iter, i);
  y_sol(iter, j2) = 0; pi(iter, i) = 0;

  PARAMETER UpperBound, LowerBound;
  UpperBound = INF; LowerBound = -INF;

* 
* SUBPROBLEM: A model in y-variables only. (and z, and p...)
*
  EQUATIONS SubObj, SubCon(i);
  POSITIVE VARIABLE p;         # Penaltx variable to ensure feasibilitx

  SubObj      .. z =E= sum( j1, c(j1) * x.L(j1) )    + 
                       sum( j2, d(j2) * y(j2) )      + BigM * p;

  SubCon(i)   ..        sum( j1, A(i,j1) * x.L(j1) ) + 
                        sum( j2, B(i,j2) * y(j2) )   + p =G= rhs(i);


  MODEL Sub /SubObj, SubCon /;

* 
* MASTER PROBLEM: A model in x-variables only (and z...)
*
  EQUATIONS MasterCut(It);

  MasterCut(cut)  .. 
        z  =G= sum( j1, c(j1) * x(j1) ) + 
               sum( j2, d(j2) * y_sol(cut, j2) ) -
               sum(i, pi(cut,i) *
                      (sum( j1, A(i,j1) * x(j1) ) +
                       sum( j2, B(i,j2) * y_sol(cut,j2) ) - rhs(i)
                      )
                   );

  z.LO = -100000;  # to bound Master Problem first time around

  MODEL Master /MasterCut /;


*
* THE ACTUAL ITERATIONS
*
  SCALAR counter / 1 /;

  x.L(j1) = 0; y.L(j2) = 0; 
  cut(Iter) = NO;

  WHILE ( UpperBound - Lowerbound > Epsilon,

    this(Iter) = yes $ (ord(Iter) eq counter);
    DISPLAY this, x.L;

    SOLVE Sub MINIMIZING z using LP;

*   Save primal and dual solution from the Sub:
    y_sol(this, j2) = y.L(j2);
    pi(this, i) = Subcon.M(i);
    UpperBound  = min(UpperBound, z.L);
    DISPLAY counter, y_sol, pi, UpperBound;

    cut(Iter)  = yes $ (ord(Iter) le counter);
    DISPLAY cut;

    IF ( UpperBound - Lowerbound > Epsilon,
      SOLVE Master MINIMIZING z using LP;
      LowerBound  = max(LowerBound, z.L);
     
      DISPLAY LowerBound;
      counter = counter + 1;
    );
  );

  DISPLAY "Benders Terminated!", counter,
           LowerBound, UpperBound, x.L, y.L;

