*
* CRR.gms: Cox-Ross-Rubinstein Stock Price simulation.
*
* GAMS Advanced Summer School. SN, June 2003.
*
$eolcom #

*
* We use the CRR stock price stochastic process:
*     dS(i,t)/S(i,t) = mu(i) * dt + sigma(i) * dz(i)
* to generate discrete-time scenarios according to:
*     Delta_S(i,t)/S(i,t) = 
*        mu(i) * Delta_t + sigma(i) * epsilon(i,t) * sqrt(Delta_t)
*
* For simplicity, we model only a single stock, i.e., drop the i index.
* In addition, however, we get a scenario index, l (the letter!).
*

  SET time /0 * 150/ ;    ALIAS(time, t);
  SET scenario / 1*1000 /; ALIAS(scenario, l);

  PARAMETER
    S_0    "Initial price"     /  100 /,
    mu     "Drift, annualized" / 0.15 /,     # 15% expected, annual return
    sigma  "Volatility"        / 0.25 /,
    Delta_t "Time intervals"   / 0.02 /,     # Weekly (about)
    epsilon(l,t) "Process increments",
    S(l,t)  "Price in scenario l at time t";
  
* Do the simulation:
  epsilon(l,t) = NORMAL(0,1);     # All increments are standard normal

  S(l,"0") = S_0;             # Start all simulations off at S_0
  LOOP(t $ (ord(t) > 1),      # Do the remaining ones (t > 0)
    S(l, t) = S(l, t-1) *
              (1 + mu * Delta_t + sigma * epsilon(l,t) * sqrt(Delta_t));
  );

* DISPLAY S;

*
* Calculate the "average path" taken by the scenarios
*
  PARAMETER avg_path(t);
  avg_path(t) = sum(l, S(l,t)) / card(l);
  DISPLAY avg_path;

*
* Calculate the "expected path" of the CRR process.
* The CRR process has a log-normal distribution (for each t),
* and its expected value at time t is NOT S_0 + mu * t, but
* S_0 + (mu - sqr(sigma)/2) * t, according to Ito's Lemma.
* Intuitively: "Down-ticks" hurt more than equivalent "up-ticks",
* so we "lose" some drift due to the volatile nature of the process.
*
  PARAMETER exp_path(t);
  exp_path(t) = exp(log(S_0) + (mu - sqr(sigma)/2) * Delta_t * (ord(t)-1));

  DISPLAY exp_path;

