


* DanCase.gms:  Model for Danish Mortgage Loans.
* See Nielsen-Zenios, "A Library of Financial Optimization Models",
* Section 7.2: "A Case Study: Danish Mortgage Loans".
*
$eolcom #
*
* RUN PARAMETERS
$set DISPL  0              # 1 if LOTS OF DISPLAY OUTPUT
$set STAGES 8              # Number of stages, 3 to 8

* The model is a multistage, stochastic optimization model
* for funding mortgage loans in Denmark, using both long
* (20- and 30-year callable bonds) and short (1-year ``Flex") bonds.
* The model is documented in "dan_mort.tex".
*
*       Copyright: Soren S. Nielsen, U. Copenhagen, June 2000.
*
* Structure of this file:
* 1. Scenario tree generation. Nodes in the trinomial tree are numbered
*    from 1 (the root, n1) upwards. Node k's successors are nodes
*    3n-1, 3n, and 3n+1. Code for generating a 3-stage, 4-stage, 5-stage,
*    and 6-stage model is included.
*
* 2. Tree parameter calculation. Using the Nielsen/Ronn interest rate model:
*    "A Two-Factor Model for the Valuation of the T-Bond Futures
*          Contract's Embedded Options", in
*          Advances in Fixed-Income Valuation Modeling and Risk Management,
*          Frank Fabozzi (ed):135--152, 1997.
*
* 3. Yield Curve Calculation. For each node, given the short and long
*    rates at the node, calculate the entire term structure.
*
* 4. Bond Data and Prices: Defines the bond universe and the relevant
*    price calculations: Callable bond prices, annuity calc's for
*    principal and interest payments, etc.
*
* 5. The optimization model. Finding the optimal sequence of portfolios
*    contingent upon the evolution of interest rates. Objective it to
*    minimize expected present value of life-time payments.
*


* SECTION 1: Scenario tree generation
*

* Specify the number of levels in the tree, the tau and delta corresponding to
* each level, and calculate the total number of nodes.
*  Tau(lvl):   time corresponding to this level, first one is 0.
*  Delta(lvl): time step between this level and next one

$if NOT "%STAGES%" == "3" $goto STAGS4
* THIS SECTION GENERATES A 3-STAGE PROGRAM WITH 40 NODES
        set node_level /level0 * level3/; alias(node_level, lvl);
        parameter taus(lvl) / level0=0, level1 = 1,
        level2=5, level3 = 30 /;
        set node /n1*n40/; alias(k, anc, node);

$label STAGS4

$if NOT "%STAGES%" == "4" $goto STAGS5
* THIS SECTION GENERATES A 4-STAGE PROGRAM WITH 121 NODES
        set node_level /level0 * level4/; alias(node_level, lvl);
        parameter taus(lvl) / level0=0, level1 = 1,
        level2=5, level3 = 10, level4 = 30 /;
        set node /n1*n121/; alias(k, anc, node);

$label STAGS5

$if NOT "%STAGES%" == "5" $goto STAGS6
* THIS SECTION GENERATES A 5-STAGE PROGRAM WITH 364 NODES
* Solve using cplex or osl
        set node_level /level0 * level5/; alias(node_level, lvl);
        parameter taus(lvl) / level0=0, level1 = 1,
        level2=2, level3 = 5, level4 = 10, level5 = 30 /;
        set node /n1*n364/; alias(k, anc, node);

$label STAGS6

$if NOT "%STAGES%" == "6" $goto STAGS7

* THIS SECTION GENERATES A 6-STAGE PROGRAM WITH 1093 NODES
* Solve using cplex or osl
        set node_level /level0 * level6/; alias(node_level, lvl);
        parameter taus(lvl) / level0=0, level1 = 1, level2 = 2,
        level3=5, level4 = 10, level5 = 20, level6 = 30 /;
        set node /n1*n1093/; alias(k, anc, node);

$label STAGS7

$if NOT "%STAGES%" == "7" $goto STAGS8

* THIS SECTION GENERATES A 7-STAGE PROGRAM WITH 3280 NODES
* Solve using cplex or osl
        set node_level /level0 * level7/; alias(node_level, lvl);
        parameter taus(lvl) / level0=0, level1 = 1, level2 = 2,
        level3 = 3,  level4 = 5, level5 = 10, level6 = 20,
        level7 = 30 /;
        set node /n1*n3280/; alias(k, anc, node);

$label STAGS8

$if NOT "%STAGES%" == "8" $goto STAGS9

* THIS SECTION GENERATES A 8-STAGE PROGRAM WITH 9841 NODES
* Solve using cplex or osl
        set node_level /level0 * level8/; alias(node_level, lvl);
        parameter taus(lvl) / level0=0, level1 = 1, level2 = 2,
        level3 = 3,  level4 = 5, level5 = 10, level6 = 15,
        level7 = 20, level8 = 30 /;
        set node /n1*n9841/; alias(k, anc, node);

$label STAGS9




  parameter horizon;   # Time corresponding to leaves
  horizon=smax(lvl, taus(lvl)); # (lvl$(ord(lvl) eq card(lvl)), taus(lvl));
  parameter deltas(lvl);
  deltas(lvl)$(taus(lvl) lt horizon)=taus(lvl+1)-taus(lvl);

  parameter num_nodes;
  num_nodes=sum(lvl, power(3, ord(lvl)-1));
  display horizon, taus, deltas, num_nodes;

* Must manually set the number of nodes to num_nodes: n1 * n(num_nodes).
  set root(node) /n1/;

* Check that tree level and number of nodes agree:
  abort $ (card(node) ne num_nodes) "Tree level and number of nodes don't agree";

* Set up the ancestor structure among nodes.
* The numbering of nodes is such that node n's successors are nodes 3n-1..3n+1;
* the root node is "n1" with ord("n1")=1.
  set ancestor(k, anc);
  ancestor(k, anc)=yes $ (3*ord(anc)-1 le ord(k) and
  3*ord(anc)+1 ge ord(k));
  parameter child(k);  # This node is child number 1, 2 or 3 from its ancestor
  child(k)=sum(ancestor(k,anc), ord(k) - 3*ord(anc) + 2);

$if "%DISPL%" == "1"    display ancestor, child;


*
* SECTION 2: Tree parameter Calculation (Nielsen/Ronn model)
*

* Each node has certain data associated with it:
*  The long and short rates at the node,
*  the node
  parameter
        long(k), short(k), # the long and short rates at the node
        level(k),          # node's distance from the root
        delta(k),          # time between this node and its successors
        tau(k),            # time corresponding to node
        prob(k),           # probability of this node
        n1(k),             # Number of up-moves leading to this node
        n2(k),             # Number of middle-moves leading to this node
        n3(k);             # Number of down-moves leading to this node

* Basic volatility data:
  parameter sigma_r, sigma_l, rho;
  sigma_r=0.32; sigma_l = 0.22; rho = 0.33; # BASE CASE FROM ROLF Apr 20 01

* Derived data used in the node calculations:
  parameter rho_temp; rho_temp=sqrt(1-sqr(rho));
  parameter a1, b1, c1, a2, b2, c2;
  a1=sqrt(3/2)*sigma_r; b1 = -a1; c1 = 0;
  a2=sigma_l*( sqrt(3/2)*rho + sqrt(1/2)*rho_temp);
  b2=sigma_l*(-sqrt(3/2)*rho + sqrt(1/2)*rho_temp);
  c2=-sigma_l*sqrt(2)*rho_temp;

* Initialize these for the root, then calculate for the rest of the tree:
  parameters short_0, long_0;
  short_0=0.044; long_0 = 0.065;

  long(root(k))=long_0;
  short(root(k))=short_0;
  level(root(k))=0;    # Distance from top of tree
  delta(root(k))=deltas("level0");
  tau(root(k))=0;
  prob(root(k))=1;
  n1(root(k))=0; n2(root(k)) = 0; n3(root(k)) = 0;

  loop(ancestor(k,anc),

        level(k)=level(anc) + 1;
        delta(k)=sum(lvl $ (ord(lvl) = level(k)+1), deltas(lvl));
        tau(k)=sum(lvl $ (ord(lvl)-1 eq level(k)), taus(lvl));
        prob(k)=1/3 * prob(anc);

        n1(k)=n1(anc) + 1 $ (child(k) eq 1);
        n2(k)=n2(anc) + 1 $ (child(k) eq 2);
        n3(k)=n3(anc) + 1 $ (child(k) eq 3);

*       Assume for this simple model that mu_s=mu_l = 0, i.e., no drift:
        short(k)=short_0 * exp((n1(k)*a1+n2(k)*b1+n3(k)*c1)*sqrt(delta(anc)));
        long(k)=long_0  * exp((n1(k)*a2+n2(k)*b2+n3(k)*c2)*sqrt(delta(anc)))

  );

  set leaves(node);    # Calculate which nodes are leaves (horizon nodes)
  leaves(k)=YES $ (tau(k) eq horizon);

$if "%DISPL%" == "1" display n1, n2, n3;
 display level, prob, long, short;
$if "%DISPL%" == "1" display tau, delta, prob, leaves;

*
* SECTION 3: Yield Curves Calculation
*
  set years /0*30/; alias(t, years);
  parameter yr(t); yr(t)=ord(t)-1; # Used to get the year corresponding to t
  parameter curve(k, t); # What the t-year yield at node k is

* Calculate node k's yield curve as a simple linear interpolation on short/long:
  curve(k,t)=(yr(t) * long(k) + (30-yr(t)) * short(k)) / 30;

  parameter DF(k); # Node k's discount factor according to root's yield curve
  DF(k) = power(1 + sum((root,t) $ (yr(t) eq tau(k)), curve(root,t)), -tau(k));
$if "%DISPL%" == "1" display DF;
*
*
* SECTION 4: Bond Data and Prices
*
*
  set bonds /Flex1 * Flex5, Fxd10-04 * Fxd10-14,
             Fxd20-04 * Fxd20-14, Fxd30-04 * Fxd30-14/;
  alias(i, bonds);
  table bond_data(i, *)  # all relevant base data. Times relative to tree start
                  issue matur  period  coupon
  Flex1           0       1     30       0.00001
  Flex2           0       2     30       0.00001
  Flex3           0       3     30       0.00001
  Flex4           0       4     30       0.00001
  Flex5           0       5     30       0.00001
  Fxd10-04        0      10     10      0.04
  Fxd10-06        0      10     10      0.06
  Fxd10-08        0      10     10      0.08
  Fxd10-10        0      10     10      0.10
  Fxd10-12        0      10     10      0.12
  Fxd10-14        0      10     10      0.14
  Fxd20-04        0      20     20      0.04
  Fxd20-06        0      20     20      0.06
  Fxd20-08        0      20     20      0.08
  Fxd20-10        0      20     20      0.10
  Fxd20-12        0      20     20      0.12
  Fxd20-14        0      20     20      0.14
  Fxd30-04        0      30     30      0.04
  Fxd30-06        0      30     30      0.06
  Fxd30-08        0      30     30      0.08
  Fxd30-10        0      30     30      0.10
  Fxd30-12        0      30     30      0.12
  Fxd30-14        0      30     30      0.14            ;

  parameter issue(i); issue(i)=bond_data(i, "issue");
  parameter matur(i); matur(i)=bond_data(i, "matur");
  parameter coupon(i); coupon(i)=bond_data(i, "coupon");
  parameter period(i); period(i)=bond_data(i, "period");
  parameter cf(i,t);   # Cash flow of bond i at time t, per 100 face value
  cf(i,t)=coupon(i) $ (yr(t) gt issue(i) and yr(t) le matur(i)) +
                         1       $ (yr(t) eq matur(i));

$if "%DISPL%" == "1" display curve, issue, matur, coupon, cf;

* Calc price of each bond at each node.

* NOTE: Check whether problem with correct accounting for coupon payments
*
*
*  Bonds have 3 different "prices":
*
*      PresentVal(k,i)  in paper:     P_i^k:  Present value of bond's cash flows if non-callable
*      MarketPrc(k,i)   in paper:     I_i^k:  Market Price of bond as callable.
*      CallPrc(k,i)     in paper:     C_i^k:  Call Price of bond, min(100, MarketPrc).

  parameter PresentVal(k,i), MarketPrc(k,i), CallPrc(k,i);

  PresentVal(k,i)=sum(t $ (yr(t) ge tau(k)),
                      cf(i,t) * (1+curve(k,t)) ** (-(yr(t) - tau(k))));

* Set prices to 0 before they are issued, and after maturity
  PresentVal(k,i) $ ((tau(k) lt issue(i)) or (tau(k) gt matur(i)))=0;

* Also truncate the price at some upper level:
  MarketPrc(k,i) =  min(PresentVal(k,i), max(1.07, 1+coupon(i)));

  CallPrc(k,i)   =  min(MarketPrc(k,i), 1.00); # Price at which issuer can call

$if "%DISPL%" == "1" display PresentVal, MarketPrc, CallPrc;


  parameter tax_rate; # Tax deduction for interest payments
  tax_rate=0.40;

  parameter trans_cost;
  trans_cost = 0.015;

*
* SECTION 5: The optimization model.
*


  positive variables          # These variables are all: PRINCIPAL amount
  x(k,i),            # Holding (face value) of bond i out of node k
  p(k,i),            # Purchases (back)
  s(k,i),            # Sales (to raise funds)
  pp(k,i), ip(k,i);  # Interest, principal payment at node k

  variables                   # These variables are all: DKK (princip. * prices)
  tot_sales(k),  # Total sales at node k (issue bonds)
  tot_purch(k),  # Total purchases at node k (buy back)
  tot_cash(k);   # Total cash out at node k
  variable z;                 # Objective value

* Implement various restrictions:
*  - can't hold a bond outside its life span (issue...maturity)
*  - can't buy at horizon or hold beyond horizon.
  s.fx(k,i) $ (tau(k) lt issue(i))=0;
  p.fx(k,i) $ (tau(k) gt matur(i))=0;
  x.fx(leaves,i)=0;
  s.fx(leaves,i)=0;

* - can't sell bonds that have a price above par:
  s.fx(k,i) $ (MarketPrc(k,i) gt 1.00)=0;

  equations initial(k,i), balance(k), bal2(k, anc,i),
  principal(k,anc,i), interest(k,anc,i), objective,
  sales_cnt(k), purch_cnt(k), cash_cnt(k);

# Individual bond balance - keep track of remaining principal,
# and sales/purchases (which may incur costs).
  initial(root(k),i) ..   s(k,i)=e= x(k,i);

  bal2(ancestor(k,anc),i)    ..  # Principal balance constraint
  x(anc,i) - pp(k,i) + s(k,i)=e= x(k,i)$(not leaves(k)) + p(k,i);

  sales_cnt(k) ..
        tot_sales(k)=e= sum(i, MarketPrc(k,i) * s(k,i));
  purch_cnt(k) ..
        tot_purch(k)=e= sum(i, CallPrc(k,i)   * p(k,i));
  cash_cnt(k) ..
        tot_cash(k)=e= sum(i, pp(k,i)  + (1-tax_rate) * ip(k,i)) +
                tot_purch(k) - tot_sales(k) +
                trans_cost * tot_sales(k) $ (not root(k));

  balance(k)$(not root(k)) ..   # no not allow prepayments
        tot_sales(k)=e= tot_purch(k);
  tot_sales.fx(root)=1000;   # Fix initial loan of 1 mill.
  tot_purch.fx(root)=0;

# Annuity calculation here:
#
# PV * (i*(1+i)^n) / ((1+i)^n-1)  is the period payment over
# n periods with period interest i that will pay off a loan of PV.
# The PMT consists of both principal and interest;
# P_j=PMT / ((1+i)^(n-j+1))  is the principal payed at the j'th period,
# j=1,...,n.

# Calculate payment per unit sold, i.e., per x(k,i), per year:
  parameter pmt(k,i),  # Total payment per unit principal, node k to succ(k)
        p_pmt(k,i),  # Principal payment period k to successor(k), per unit
        i_pmt(k,i);  # Ditto interest payment


  loop((ancestor(k,anc), i) $ ((matur(i) gt tau(anc)) and (MarketPrc(anc,i) gt 0)),
            pmt(anc,i) = coupon(i)*power(1+coupon(i), period(i)-tau(anc)) /
                         (power(1+coupon(i),period(i)-tau(anc)) -1);

          p_pmt(anc,i) = sum(t $ (yr(t) ge issue(i) and yr(t) lt matur(i) and
                                  yr(t) ge tau(anc) and yr(t) lt tau(k)),
                                  # Calculate principal payment at year
                                  1/power(1+coupon(i), period(i) - yr(t))
                             ) * pmt(anc,i);

          i_pmt(anc,i) = sum(t $ (yr(t) ge issue(i) and yr(t) lt matur(i) and
                                  yr(t) ge tau(anc) and yr(t) lt tau(k)),
                                  # Calculate principal payment at year
                                  1 - 1/power(1+coupon(i), period(i) - yr(t))
                             ) * pmt(anc,i);
  );
$if "%DISPL%" == "1" display pmt, p_pmt, i_pmt;

  principal(ancestor(k,anc),i) .. # principal payment at node k, no prepay
          pp(k,i)=e=  p_pmt(anc,i) * x(anc,i);
          pp.fx(root,i)=0;

  interest(ancestor(k,anc),i) ..  # interest payment at node k
          ip(k,i)=e= i_pmt(anc,i) * x(anc,i);
          ip.fx(root,i)=0;

  objective ..    # Minimize PV of total expected payments, except at root.
                  # Discount by the initial (root) yield curve, DF(k)
         z =e= sum(k $ (not root(k)), prob(k) * tot_cash(k) * DF(k));

  model Danish /all/;


  option iterlim=999999, reslim = 10000, decimals = 6;
  option limrow = 0, limcol = 0, solprint = off;

  option lp=cplex;  # or cplex, for the larger models
  solve Danish minimizing z using LP;

* Display key data and results:
  display level, long, short;
  display x.l, p.l, s.l;