
* Newsvendor problem, for use with the Advanced Summer School.
*
$eolcom #

********************************************************************************
*                                                                              *
*  SETTING UP THE PROBLEM'S DATA AND INITIAL CALCULATIONS:                     *
*                                                                              *
********************************************************************************

* In the "real" Newsvendor problem, any demand is possible to occur, but GAMS
* can't handle an infinite amount of possibilities. Therefore,
* the set "support" represents the set of possible demands for papers realized.
* For simplicity, the number 0 is excluded (so we can use the "ord" function directly
* to get from the set element "4" to the value 4)

 SET support / 1 * 100 /;   ALIAS( support, i, j );

 PARAMETER p(i) "Probility that i papers will be demanded";
 PARAMETER lambda / 20.5 /;

* Calculate p(i) as a truncated Poisson-distribution with parameter lambda:  
* p(i) = exp(-lambda) * lambda^i / x!   scaled to sum to 1.

 LOOP( i,                             # Runs through i = "1", "2", ...
   IF (ord(i) = 1,                    # Calculate the first one from scratch
     p("1") = exp(-lambda) * lambda;
   ELSE                               # Calculate based on the previous one
     p(i) = p(i-1) * lambda / ord(i);
   )
 );
 p(i) = p(i) / sum(j, p(j));    # Normalize so the p(i)'s sum to 1
 DISPLAY p;

* Done calculating p(i).

 PARAMETERS
    c_s   "Price at which papers are sold"             / 35 /,
    c_p   "Price at which papers are purchased"        / 20 /,
    c_f   "Price for firh-wrapping (unsold) papers"    /  2 /,
    N     "Max number of papers that can be purchased" / 40 /;

 PARAMETER F(i,j) "Profit if i papers are purchased, and j are demanded";
 F(i,j) = c_s * min(ord(i), ord(j)) + c_f * max(0, ord(i)-ord(j)) - c_p * ord(i);
 DISPLAY F; 

********************************************************************************
*                                                                              *
*  STRAIGHT-FORWARD CALCULATION OF OPTIMAL VALUE:                              *
*                                                                              *
********************************************************************************

 PARAMETER G(i) "Expected profit if i papers are purchased";

* Calculate G(i), leave the profit equal to zero for i's greater than N:
 G(i) $ (ord(i) <= N) = sum(j $ (ord(j) <= ord(j)), p(j) * F(i,j) );
 DISPLAY G;

* Find the highest possible expected profit, and the associated number (i_hat) to 
* purchase:
 PARAMETER i_hat, max_prof;
 max_prof = SMAX(i, G(i));

 i_hat = SMIN(i $ (G(i) = max_prof), ord(i));
 DISPLAY i_hat, max_prof;

********************************************************************************
*                                                                              *
*  END STRAIGHT-FORWARD CALCULATION OF OPTIMAL VALUE.                          *
*                                                                              *
********************************************************************************







********************************************************************************
*                                                                              *
*  NEWSVENDOR PROBLEM AS A TWO-STAGE, STOCHASTIC PROGRAM.                      *
*                                                                              *
********************************************************************************

* We now formulate the problem as a two-stage, linear stochastic program (SP).
* The first-stage decision is: How many papers to purchase, represented by x.
* The second-stage decision is: How many papers to sell, represented by z(s).
* The index s is the "scenario", that is, how many papers are demanded.
*
* Notice the two-stage decision process (two decision points):
* 1. Purchases in the morning, then wait and see how many are demanded/sold,
* 2. Decide how many papers to sell as fish wrapping, in this case a trivial decision.

* The scenario set (demands) is just the "support" set defined above, but we'd
* like to use the index "s" over scenarios:
  ALIAS( support, s );

  POSITIVE VARIABLES x      "How many papers to purchase",
                     z(s)   "How many papers are sold";

* Constraints on the first-stage variable(s):
  EQUATION max_prch "Max number to purchase";
  max_prch ..  x =L= N;

* Constraints on the second-stage variables:
  EQUATIONS Cnst2_a(s), Cnst2_b(s);

  Cnst2_a(s)  .. z(s) =L= x;         # Can't sell more than we have purchased
  Cnst2_b(s)  .. z(s) =L= ord(s);    # Under scenario s, ord(s) papers are demanded


* Calculate the objective:
  VARIABLES prof(s)     "Profit under each scenario",
            Exp_Prof    "Expected profit of the whole day";

  EQUATIONS Prof_Def(s), Exp_Prof_Def;

  Prof_Def(s)  .. prof(s) =E= c_s * z(s) + c_f * (x-z(s)) - c_p * x;
  Exp_Prof_Def .. exp_prof =E= sum(s, p(s) * prof(s) );

  MODEL TWO_STAGE /ALL/;
  SOLVE two_stage MAXIMIZING exp_prof USING LP;


